/*
 * PNG Foo -- screw with basic PNG images; Brandon Enright; 2008-05-31
 *
 * This program was hacked together specifically to solve the
 * Forensics 500 PNG image challenge given by Kenshoto (www.kenshoto.com)
 * for the DEFCON CTF 2008 qualification contest.
 *
 * This program is a modified version of flipchannels which itself is a
 * hack of a program out of mandrakesoft.com.
 *
 * Copyrights below.
 *
 * Guillaume Cottenceau (gc at mandrakesoft.com)
 *
 * Small modification for R/B channel flippage by MHW.
 *
 * Small modification for f500 by Brandon Enright <bmenrigh@ucsd.edu>.
 *
 * Everything else is Copyright 2002 MandrakeSoft.
 *
 * This software may be freely redistributed under the terms of the GNU
 * public license.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#define PNG_DEBUG 3
#include <png.h>

void abort_(const char * s, ...)
{
    va_list args;
    va_start(args, s);
    vfprintf(stderr, s, args);
    fprintf(stderr, "\n");
    va_end(args);
    abort();
}

int x, y;

int width, height;
png_byte color_type;
png_byte bit_depth;

png_structp png_ptr;
png_infop info_ptr;
int number_of_passes;
png_bytep * row_pointers;

/* **** These have been added for f500 **** */
#define COUNT 248 * 563 * 4 /* How many bytes we have (not secure) */
png_byte randomb[COUNT];     /* An array for our random stream */


void read_png_file(char* file_name) {
  unsigned char header[8]; /* 8 is the maximum size that can be checked */

  /* open file and test for it being a png */
  FILE *fp = fopen(file_name, "rb");
  if (!fp)
    abort_("[read_png_file] File %s could not be opened for reading",
	   file_name);
  fread(header, 1, 8, fp);
  if (png_sig_cmp(header, 0, 8))
    abort_("[read_png_file] File %s is not recognized as a PNG file",
	   file_name);

  /* initialize stuff */
  png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  
  if (!png_ptr)
    abort_("[read_png_file] png_create_read_struct failed");

  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr)
    abort_("[read_png_file] png_create_info_struct failed");
  
  if (setjmp(png_jmpbuf(png_ptr)))
    abort_("[read_png_file] Error during init_io");
  
  png_init_io(png_ptr, fp);
  png_set_sig_bytes(png_ptr, 8);
  
  png_read_info(png_ptr, info_ptr);
  
  width = info_ptr->width;
  height = info_ptr->height;
  color_type = info_ptr->color_type;
  bit_depth = info_ptr->bit_depth;
  
  number_of_passes = png_set_interlace_handling(png_ptr);
  png_read_update_info(png_ptr, info_ptr);
  
  /* read file */
  if (setjmp(png_jmpbuf(png_ptr)))
    abort_("[read_png_file] Error during read_image");
  
  row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * height);
  for (y=0; y<height; y++)
    row_pointers[y] = (png_byte*) malloc(info_ptr->rowbytes);
  
  png_read_image(png_ptr, row_pointers);
  
  fclose(fp);
}


void write_png_file(char* file_name) {
  /* create file */
  FILE *fp = fopen(file_name, "wb");
  if (!fp)
    abort_("[write_png_file] File %s could not be opened for writing",
	   file_name);

  /* initialize stuff */
  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  
  if (!png_ptr)
    abort_("[write_png_file] png_create_write_struct failed");
  
  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr)
    abort_("[write_png_file] png_create_info_struct failed");
  
  if (setjmp(png_jmpbuf(png_ptr)))
    abort_("[write_png_file] Error during init_io");
  
  png_init_io(png_ptr, fp);
  
    /* write header */
  if (setjmp(png_jmpbuf(png_ptr)))
    abort_("[write_png_file] Error during writing header");
  
  png_set_IHDR(png_ptr, info_ptr, width, height,
	       bit_depth, color_type, PNG_INTERLACE_NONE,
	       PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
  
  png_write_info(png_ptr, info_ptr);
  
    /* write bytes */
  if (setjmp(png_jmpbuf(png_ptr)))
    abort_("[write_png_file] Error during writing bytes");
  
  png_write_image(png_ptr, row_pointers);
  
  /* end write */
  if (setjmp(png_jmpbuf(png_ptr)))
    abort_("[write_png_file] Error during end of write");
  
  png_write_end(png_ptr, NULL);
  
  /* cleanup heap allocation */
  for (y=0; y<height; y++)
    free(row_pointers[y]);
  free(row_pointers);
  
  fclose(fp);
}


void process_file(void) {

  /* Okay here is where the Forensics 500 magic happens
   *
   * There are a few things you'd want to do with this image to
   * solve the problem.  Many of those things will lead to dead ends
   * but I feel that they are required to make the logical deductions
   * necessary to finally arrive at the solution.
   *
   * I'm going to leave most of the tasks commented out but the code
   * to do them will be more-or-less left intact.
   *
   * 1) Remove the alpha mask
   * 2) Isolate a color channel (or the alpha mask)
   * 3) Replace a color channel with another or swap channels
   * 4) Produce a new image of just the hidden data (leaves space at the end)
   * 5) Muck around with the random data as a byte stream
   * 6) Output channel data
   * 7) Output pixel data
   * 8) Higlight pixels
   * 9) Do basic math on higlighted pixels
   *
   * Realistically, you only need to do 1, 7, 8, and 9 to solve the problem.
   *
   * The PNG format stores the image data in rows with row filters.
   * libpng hides the row filters from us and the bytes in the rows
   * are already post-filter.  The channel order is RGBA.
   *
   * Here is how to access each channel:
   * R: *(*(row_pointers + i) + y)
   * G: *(*(row_pointers + i) + y + 1)
   * B: *(*(row_pointers + i) + y + 2)
   * A: *(*(row_pointers + i) + y + 3)
   *
   * The image is fully transparent at alpha 0x00 and solid at alpha 0xff.
   *
   * If you want to treat the image as pixel data rather than channel data
   * you can convert the channels into a 24-bit number like so:
   *
   * val = ((*(*(row_pointers+i)+y) << 16) +
   *       (*(*(row_pointers+i)+y+1) << 8) +
   *       (*(*(row_pointers+i)+y+2)));
   *
   * For the the challenge, the marker pixel value is 965378
   *
   */

  int i; /* the row counter */

  long val;

  int oldx;
  int soli = 0;
  char solution[50]; /* Security be damned */

  int count = 0;
  png_byte xor = 0;

  /* Make sure we have a RGBA image */
  if (info_ptr->color_type != PNG_COLOR_TYPE_RGBA)
    abort_("[process_file] color_type of input file must be PNG_COLOR_TYPE_RGBA (is %d)", info_ptr->color_type);
  
  oldx = 0;
  /* here is the main loop through the image data */
  for (i = 0; i < height; i++){
    for (y = 0; y < width * 4; y += 4){
      png_byte tmp; /* used for swapping */
      

      /* 1) Remove alpha mask 
       * Simply max out the alpha channel:
       *
       * *(*(row_pointers+i)+y+3) = 0xff;
       *
       * Don't forget that if you do this, later operations on the alpha
       * will see 0xff only...
       */


      /* 2) Isolate a channel
       * Set the alpha to 0xff and the other channels to 0x00
       * Here is how to see green only:
       *
       * *(*(row_pointers+i)+y + 0) = 0x00;
       * *(*(row_pointers+i)+y + 2) = 0x00;
       * *(*(row_pointers+i)+y + 3) = 0xff;
       */

      
      /* 3) Swap color channels
       * To swap red and blue use the tmp val (the iPhone needs this operation)
       *
       * tmp = *(*(row_pointers+i)+y + 0);
       * *(*(row_pointers+i)+y + 0) = *(*(row_pointers+i)+y + 2);
       * *(*(row_pointers+i)+y + 2) = tmp;
       *
       */


      /* 4a) Produce a new image with just the random data
       * This is done in two parts, the first part is to record
       * all of the random data in the randomb[] array and then later
       * put it back in the image.  Here is the first part:
       *
       * if (*(*(row_pointers+i)+y+3) == 0x00) {
       *   randomb[count++] = *(*(row_pointers+i)+y + 0);
       *   randomb[count++] = *(*(row_pointers+i)+y + 1);
       *   randomb[count++] = *(*(row_pointers+i)+y + 2);
       * }
       *
       */


      /* 5a) Muck around with the random data
       * To use the random data you first need to record it which is done
       * in step 4a.
       */


      /* 6) Output channel data
       * This is useful if you want to use gnuplot or another utility
       * to mess around with the channels or colors
       *
       *  printf("%d %d %d\n", 
       *         (*(*(row_pointers+i)+y) & 0xff),
       *         (*(*(row_pointers+i)+y+1) & 0xff),
       *         (*(*(row_pointers+i)+y+2) & 0xff));
       *
       */

	
      /* 7) Output pixel data
       * This requires that you first compute the pixel value
       *
       * val = ((*(*(row_pointers+i)+y) << 16) +
       *       (*(*(row_pointers+i)+y+1) << 8) +
       *       (*(*(row_pointers+i)+y+2)));
       *
       * printf("%d\n", val);
       *
       */

	
      /* 8) Highlight a pixel value
       * Using the pixel value computed in 7, you can highlight the
       * random-looking pixels as white and make the rest of them black
       *
       * if (*(*(row_pointers+i)+y + 3) == 0x00) {
       *   if (val == 965378) {
       *     *(*(row_pointers+i)+y + 0) = 0xff;
       *     *(*(row_pointers+i)+y + 1) = 0xff;
       *     *(*(row_pointers+i)+y + 2) = 0xff;
       *   }
       *   else {
       *     *(*(row_pointers+i)+y + 0) = 0x00;
       *     *(*(row_pointers+i)+y + 1) = 0x00;
       *     *(*(row_pointers+i)+y + 2) = 0x00;
       *   }
       * }
       *
       */


      /* 9) Do basic math on the highligted pixels
       * The following will compute the solution to f500
       */
	
      val = ((*(*(row_pointers+i)+y) << 16) +
             (*(*(row_pointers+i)+y+1) << 8) +
             (*(*(row_pointers+i)+y+2)));
      
      if (val == 965378) {
	if (oldx != 0)  {
	  printf("Y coord: %-3d; diff: %-3d; char: %c\n",
		 i,
		 i - oldx,
		 'A' + (i - oldx - 1));
 	
	  /* The real solution is -2, not -1 */
	  solution[soli++] = 'A' + (i - oldx - 2);
	  solution[soli] = '\0';
	}

	oldx = i;
      }

    } /* end of column loop */
  }  /* end of row loop */
  

  /* 5b) muck around with the random data
   * If you want to mess with the random data you can do so in this loop.
   * This is the XOR chain example
   *
   * count = 0;
   * while (count < COUNT) {
   *   xor = (xor ^ randomb[count]);
   *   randomb[count++] = xor;
   * }
   */
  
  /* 4b) Save hidden mask data
   * This will save back all the masked data to the image.  It relies on
   * the tail end of randomb[] to be constant (it is in the data section)
   *
   * count = 0;
   * for (i = 0; i < height; i++){
   *   for (y = 0; y < width * 4; y += 4){
   *     *(*(row_pointers+i)+y + 0) = randomb[count++];
   *     *(*(row_pointers+i)+y + 1) = randomb[count++];
   *     *(*(row_pointers+i)+y + 2) = randomb[count++];
   *   }
   * }
   */
  
  printf("\n"); /* for good measure */
  /* print our solution */
  printf("Solution: %s\n", solution);

} /* end of process_file */


int main(int argc, char **argv) {
  if (argc != 3)
    abort_("Usage: program_name <file_in> <file_out>");
  
  read_png_file(argv[1]);
  process_file(); /* The f500 stuff */
  write_png_file(argv[2]); /* save your changes */
  
  return 0;
}
